;       Compile with WLA-6510 (http://www.hut.fi/~vhelin/wla.html)

.DEFINE startaddr       $1000

.DEFINE timing          $0f5a           ; 2x = $2663, 4x = $1331, 5x = $0f5a

.DEFINE freq1           $02             ; osc1 base frequency
.DEFINE freqh1          $03

.DEFINE freq2           $04             ; osc2 base frequency
.DEFINE freqh2          $05

.DEFINE acc             $06             ; $06-$0b: multiply
.DEFINE aux             $08
.DEFINE ext             $0a

.DEFINE keybase         $0c             ; first note of keyboard
.DEFINE editparam       $0d             ; current edit parameter

.DEFINE lastkey         $0e             ; last key down
.DEFINE temp            $0f
.DEFINE scr             $10
.DEFINE scrh            $11

.DEFINE ccolor          $12             ; cursor/parameter color
.DEFINE flash           $13             ; flashing cursor
.DEFINE currentpatch    $14             ; $10 bytes current patch

.DEFINE bend            $fc
.DEFINE bendh           $fd

.DEFINE tmpfreq         $fe
.DEFINE tmpfreqh        $ff


;-----  color palette

.DEFINE color1          $0f             ; text color
.DEFINE color2          $06             ; parameter/value key color
.DEFINE color3          $0b             ; current value color
.DEFINE color4          $01             ; selected parameter line text color


.MEMORYMAP
DEFAULTSLOT 0
SLOTSIZE $ffff
SLOT 0 $0000
.ENDME

.ROMBANKMAP
BANKSTOTAL 1
BANKSIZE $ffff
BANKS 1
.ENDRO

.BANK 0

.ORG $0
.SECTION "BANKHEADER"
        .dw     startaddr
.ENDS


.ORG startaddr
.SECTION "main" FORCE

Go:
        sei


;-----  set up interrupts

        lda     #irq&255
        sta     $fffe
        lda     #irq>>8
        sta     $ffff

        lda     #nmi&255
        sta     $fffa
        lda     #nmi>>8
        sta     $fffb

        lda     #timing&255
        sta     $dc04
        lda     #timing>>8
        sta     $dc05

        lda     #$35
        sta     $01


;-----  set up variables

        lda     #$24
        sta     keybase

        lda     #$ff
        sta     editparam


;-----  screen

        ldx     #$00
        stx     $d020
        stx     $d011
        stx     $d021

_sclp
        lda.w   screen,x
        sta.w   $0400,x
        lda.w   screen+$0100,x
        sta.w   $0500,x
        lda.w   screen+$0200,x
        sta.w   $0600,x
        lda.w   screen+$0300,x
        sta.w   $0700,x

        lda     #color1
        sta     $d800,x                 ; set text color
        sta     $d900,x
        sta     $da00,x
        sta     $db00,x

        inx
        bne     _sclp

        lda     #$18
        sta     $d018
        lda     #$08
        sta     $d016
        lda     #color3
        sta     $d022
        lda     #color2
        sta     $d023

;-----  load patch 1

        lda     #$01
        jsr     loadpatch

        lda     #$5b
        sta     $d011


;-----  main loop

        cli

mainloop:
        jmp     mainloop


;-----  synth irq

irq:
        pha
        txa
        pha
        tya
        pha

        jsr     readkeyboard
        jsr     update
        jsr     pitchback

        inc     flash
        lda     flash
        lsr
        lsr
        lsr
        lsr
        and     #$07
        tax
        lda.w   flashtab,x
        sta     $d024

        lda     $dc0d

        pla
        tay
        pla
        tax
        pla

nmi:
        rti


;-----  load patch (a = 1-9)

loadpatch:
        pha
        ora     #$30                    ; write patch number on screen
        sta     $07cf-8
        lda     #$2d
        sta     $07d8-8
        pla

        asl
        asl
        asl
        asl
        tay

        ldx     #$00

_loadlp
        lda.w   patchdata-$10,y
        sta     currentpatch,x
        iny
        inx
        cpx     #$10
        bne     _loadlp

        lda     #$00
        ldx     #$1f
_inilp
        sta.w   $d400,x                 ; SIZEOP: combine loops
        dex
        bpl     _inilp

        stx     lastkey

        lda     #$00
        sta     freq1
        sta     freq2
        sta     freqh1
        sta     freqh2

        jmp     snapreset


;-----  save patch (a = 1-9)

savepatch:
        pha
        ora     #$30                    ; write patch number on screen
        sta     $07cf-8
        sta     $07d8-8
        pla

        asl
        asl
        asl
        asl
        tay

        ldx     #$00

_savelp
        lda     currentpatch,x
        sta.w   patchdata-$10,y
        iny
        inx
        cpx     #$10
        bne     _savelp

        rts


;-----  note on (a = 0-127)

noteon:
        ldx     #$00
        stx     $d400
        stx     $d401
        stx     $d407
        stx     $d408
        stx     bend
        stx     bendh

        pha

        clc                             ; osc 1 transpose
        ldx     currentpatch+$02
        adc.w   osctranspose-1,x

        asl
        tax
        lda.w   freqtab,x
        sta     freq1
        lda.w   freqtab+1,x
        sta     freqh1

        pla

        clc                             ; osc 2 transpose
        ldx     currentpatch+$05
        adc.w   osctranspose-1,x

        asl
        tax
        lda.w   freqtab,x
        sta     freq2
        lda.w   freqtab+1,x
        sta     freqh2

        ldx     currentpatch+$01        ; osc 1 envelope
        lda.w   envelopesad-1,x
        sta     $d405
        lda.w   envelopessr-1,x
        sta     $d406

        ldx     currentpatch+$04        ; osc 2 envelope
        lda.w   envelopesad-1,x
        sta     $d40c
        lda.w   envelopessr-1,x
        sta     $d40d

        ldx     currentpatch+$0b        ; filter envelope
        lda.w   envelopesad-1,x
        sta     $d413
        lda.w   envelopessr-1,x
        sta     $d414

        ldx     currentpatch+$0e        ; filter type
        lda.w   filtertypes-1,x
        sta.w   $d418

        ldx     currentpatch+$0f        ; filter resonance
        lda.w   filterresos-1,x
        sta.w   $d417

        ldx     currentpatch+$08        ; lfo rate
        lda.w   lforates-1,x
        sta     $d40e
        lda     #$00
        sta     $d40f

        lda     #$11                    ; lfo/mod gate off
        sta     $d412

        ldx     currentpatch            ; osc 1 waveform, gate on
        lda.w   osc1waveforms-1,x
        ora     #$01
        sta     $d404

        ldx     currentpatch+$03        ; osc 2 waveform, gate on
        lda.w   osc2waveforms-1,x
        ora     #$01
        sta     $d40b

        rts


;-----  update synth modulation etc.

update:
        ldx     currentpatch+$09        ; lfo vibrato
        dex
        stx     acc
        lda     #$00
        sta     acc+1

        ldy     #$00                    ; read and sign-extend lfo value
        lda     $d41b                   
        eor     #$80
        sta     aux
        bpl     _lnos1
        dey
_lnos1
        sty     aux+1

        jsr     multiply

        ldx     #$03

_sclfov
        lda     ext+1                   ; scale down lfo vibrato
        cmp     #$80
        ror
        ror     ext
        sta     ext+1
        dex
        bne     _sclfov

        lda     freq1                   ; osc 1 frequency
        clc
        adc     ext
        sta     tmpfreq ;$d400
        lda     freqh1
        adc     ext+1
        sta     tmpfreqh ;$d401

        jsr     addbend
        lda     tmpfreq
        sta     $d400
        lda     tmpfreqh
        sta     $d401

        lda     freq2                   ; osc 2 frequency
        clc
        adc     ext
        sta     acc
        lda     freqh2
        adc     ext+1
        sta     acc+1

        ldx     currentpatch+$06        ; osc 2 detune
        txa
        asl                             ; scale up detune
        asl
        clc
        adc     acc
        sta     tmpfreq ;$d407
        lda     acc+1
        adc     #$00
        sta     tmpfreqh ;$d408

        jsr     addbend
        lda     tmpfreq
        sta     $d407
        lda     tmpfreqh
        sta     $d408

        lda     currentpatch+$0a        ; lfo pwm
        sta     acc
        lda     #$00
        sta     acc+1

        lda     $d41b
        sta     aux
        lda     #$00
        sta     aux+1

        jsr     multiply
        
        lda     ext+1                   ; scale down pulse width modulation
        cmp     #$80
        ror
        ror     ext
        sta     ext+1

        ldx     currentpatch+$07        ; base pulse width
        lda     ext
        sta     $d402
        sta     $d409
        lda.w   pulsewidths-1,x
        clc
        adc     ext+1                   
        sta     $d403      
        sta     $d40a


        ldx     currentpatch+$0d        ; filter modulation amount
        ldy     #$00
        lda.w   filtermods-1,x
        sta     acc
        bpl     _lnos3
        dey
_lnos3
        sty     acc+1

        lda     $d41c                   ; read filter envelope value
        sta     aux
        lda     #$00
        sta     aux+1

        jsr     multiply

       
        ldx     currentpatch+$0c        ; base filter cutoff
        lda     ext+1
        clc
        adc.w   filtercutoffs-1,x
        sta     $d416

        rts


;-----  note off

noteoff:
        ldx     currentpatch            ; osc 1 gate off
        lda.w   osc1waveforms-1,x
        sta     $d404

        ldx     currentpatch+$03        ; osc 2 gate off
        lda.w   osc2waveforms-1,x
        sta     $d40b

        lda     #$10                    ; lfo/mod gate off
        sta     $d412

        rts


;-----  multiply acc * aux -> ext

multiply:
        lda     #$00
        sta     ext
        sta     ext+1

_mbit
        lsr     aux+1
        ror     aux
        bcc     _mnext

        lda     acc
        clc
        adc     ext
        sta     ext
        lda     acc+1
        adc     ext+1
        sta     ext+1

_mnext
        asl     acc
        rol     acc+1

        lda     aux+1
        bne     _mbit
        lda     aux
        bne     _mbit

        rts



;-----  pitch bends

addbend:
        lda     tmpfreq
        clc
        adc     bend
        sta     tmpfreq
        lda     tmpfreqh
        adc     bendh
        sta     tmpfreqh
        rts

pitchback:
        lda     bendh
        bne     _notz1
        lda     bend
        bne     _notz1
        rts

_notz1
        lda     bendh
        bmi     _backup
_backdown
        lda     bend
        clc
        adc     #$f0
        sta     bend
        lda     bendh
        adc     #$ff
        sta     bendh

        rts
_backup
        lda     bend
        clc
        adc     #$10
        sta     bend
        lda     bendh
        adc     #$00
        sta     bendh

        rts


pitchup:
        lda     bend
        clc
        adc     #$20
        sta     bend
        lda     bendh
        adc     #$00
        sta     bendh

        jmp     _ctrldone

pitchdown:
        lda     bend
        clc
        adc     #$e0
        sta     bend
        lda     bendh
        adc     #$ff
        sta     bendh

        jmp     _ctrldone


;-----  read keyboard

readkeyboard:

;-      control keys

        ldy     #$00
_nextctrl
        lda.w   controlkeylist,y        ; check control key
        beq     _ctrldone
        sta     $dc00
        lda     $dc01
        and.w   controlkeylist+1,y
        bne     _noctrlkey

        lda.w   controlkeylist+2,y      ; octave change or edit?
        bpl     _noteditmode

        sta     lastkey                 ; no last key (>$80)
        jmp     _editmode

_noteditmode
        cmp     #$01                    ; pitch bends
        beq     pitchdown
        cmp     #$02
        beq     pitchup

        cmp     keybase
        beq     _noctrlkey

        sta     keybase                 ; set new keyboard octave
        lda     #$ff
        sta     lastkey
        rts

_noctrlkey
        iny
        iny
        iny
        bne     _nextctrl               ; always branch
_ctrldone


;-      musical keyboard


        ldy     #$1f*$03                ; priority to highest key

_nextkey
        lda.w   keyboardlist,y
        sta     $dc00
        lda     $dc01
        and.w   keyboardlist+1,y
        beq     _keydown
_oknx
        dey
        dey
        dey
        bpl     _nextkey

_keysdone
        sty     lastkey                 ; no last key (>$1f)
        jmp     noteoff                 ; no keys, gate off, rts there

_keydown
        lda.w   keyboardlist+2,y
        cmp     lastkey                 ; check if key already down
        beq     _notrig

        sta     lastkey

        clc
        adc     keybase
        pha
        jsr     noteoff
        pla
        jmp     noteon


;-      check edit mode keys

_editmode
        jsr     noteoff

        ldy     #$00
_nextedit
        lda.w   editkeylist,y           ; scan edit keys
        beq     _editdone
        sta     $dc00
        lda     $dc01
        and.w   editkeylist+1,y
        bne     _noeditk

        ldx.w   editkeylist+2,y         
        bmi     _editchange             ; negative are parameter change

        cpx     editparam               ; select new edit parameter
        beq     _editdone

        txa
        pha

        ldx     editparam
        cpx     #$10
        bcs     _nocpr

        ldy     #color1
        jsr     hiliteparam             ; clear current parameter color

        inc     editparam               ; remove cursor
        jsr     drawparamline           

_nocpr
        lda     $07cf-8                  ; remove bottom line cursor
        and     #$3f
        sta     $07cf-8
        lda     $07d8-8
        and     #$3f
        sta     $07d8-8

        pla
        tax
        stx     editparam
        cpx     #$10
        bcs     _patchoption

        ldy     #color4
        jsr     hiliteparam             ; hilite new parameter
        jsr     drawparamline           ; draw cursor

        jmp     snap


_noeditk
        iny
        iny
        iny
        bne     _nextedit               ; always branch

_patchoption
        ldy.w   botlinecur-$10,x        ; bottom line option cursor
        lda     $07cf-8,y
        ora     #$c0
        sta     $07cf-8,y
        jmp     snap

_notrig
_editdone
        rts

_editchange
        ldy     editparam               ; editparam negative if no edit
        bmi     _editdone

        cpy     #$10
        beq     ctrl_savepatch
        cpy     #$11
        beq     ctrl_loadpatch

        txa
        and     #$0f                    ; new value for parameter
        cmp.w   currentpatch,y
        beq     _editdone               ; no snap if already same value

        sta.w   currentpatch,y          ; store parameter change

        lda     #$2d                    ; clear store-column to indicate
        sta     $07d8-8                 ; that the sound has been changed

        ldx     editparam
        jsr     drawparamline           ; draw parameter line

        jmp     snap


;-      ctrl_loadpatch

ctrl_loadpatch:
        txa
        and     #$0f
        jmp     loadpatch


;-      ctrl_savepatch

ctrl_savepatch:
        txa
        and     #$0f
        jsr     savepatch


;-      snap and reset edit parameter, screen colors, draw sound parameters

snapreset:
        lda     #$ff                    ; reset edit parameter
        sta     editparam

        ldx     #$0f
_drawlp
        jsr     drawparamline
        dex
        bpl     _drawlp
        ;
        ;
;-      snap sound to confirm edits

snap:
        ldx     #$02
_snl
        lda     #$00
        jsr     tick
        lda     #$0f
        jsr     tick
        dex
        bne     _snl

        ldx.w   currentpatch+$0e        ; filter type
        lda.w   filtertypes-1,x
        sta.w   $d418

        rts

tick:
        sta     $d418

        ldy     #$40
_dl1
        dey
        bpl     _dl1

        rts


;-----  hilite parameter line x using color y

hiliteparam:
        lda.w   parline,x
        sta     scr
        lda.w   parlineh,x
        eor     #$dc
        sta     scrh

        tya
        ldy     #$27
_hililp
        sta     (scr),y
        dey
        bpl     _hililp

        rts


;-----  draw parameter line x

drawparamline:
        stx     temp
        lda     #$40
        cpx     editparam
        bne     _noteditn
        lda     #$c0
_noteditn
        sta     ccolor

        lda.w   parline,x               ; parameter line
        sta     scr
        lda.w   parlineh,x
        sta     scrh

        ldy     #$27
_wipl
        lda     (scr),y                 ; reverse off all
        and     #$3f
        sta     (scr),y
        dey
        bne     _wipl

        lda     currentpatch,x          ; get parameter value
        asl
        asl
        clc
        adc     #$01
        tay

        ldx     #$03
_revl
        lda     (scr),y                 ; reverse current parameter value
        ora     ccolor
        sta     (scr),y
        iny
        dex
        bne     _revl

        ldx     temp
        rts


;-----  misc data

freqtab:
        .INCBIN "freqtab.bin"

osc1waveforms:
        .DB     $10,$20,$40,$80,$30,$50,$70,$00,$08

osc2waveforms:
        .DB     $10,$20,$40,$80,$12,$22,$42,$14,$08

osctranspose:
        .DB     $e8,$ef,$f4,$fb,$00,$07,$0c,$11,$18

pulsewidths:
        .DB     $00,$01,$02,$03,$04,$05,$06,$07,$08

lforates:
        .DB     $08,$10,$18,$20,$30,$40,$60,$80,$c0

envelopesad:
        .DB     $00,$09,$0a,$05,$48,$cd,$b5,$3a,$0b

envelopessr:
        .DB     $d0,$95,$08,$05,$85,$49,$15,$1a,$0b

filtercutoffs:
        .DB     $00,$10,$20,$30,$40,$50,$60,$70,$7f

filtermods:
        .DB     $00,$10,$20,$30,$40,$50,$60,$70,$7f

filtertypes:
        .DB     $9f,$af,$cf,$bf,$df,$ef,$ff,$98,$b8

filterresos:
        .DB     $0b,$1b,$3b,$5b,$7b,$9b,$bb,$db,$fb


;-----  keyboard map

.DEFINE ka0     $ff-$01
.DEFINE ka1     $ff-$02
.DEFINE ka2     $ff-$04
.DEFINE ka3     $ff-$08
.DEFINE ka4     $ff-$10
.DEFINE ka5     $ff-$20
.DEFINE ka6     $ff-$40
.DEFINE ka7     $ff-$80

.DEFINE kb0     $01
.DEFINE kb1     $02
.DEFINE kb2     $04
.DEFINE kb3     $08
.DEFINE kb4     $10
.DEFINE kb5     $20
.DEFINE kb6     $40
.DEFINE kb7     $80


editkeylist:
        .DB     ka1,kb2,$00     ;A
        .DB     ka3,kb4,$01     ;B
        .DB     ka2,kb4,$02     ;C
        .DB     ka2,kb2,$03     ;D
        .DB     ka1,kb6,$04     ;E
        .DB     ka2,kb5,$05     ;F
        .DB     ka3,kb2,$06     ;G
        .DB     ka3,kb5,$07     ;H
        .DB     ka4,kb1,$08     ;I
        .DB     ka4,kb2,$09     ;J
        .DB     ka4,kb5,$0a     ;K
        .DB     ka5,kb2,$0b     ;L
        .DB     ka4,kb4,$0c     ;M
        .DB     ka4,kb7,$0d     ;N
        .DB     ka4,kb6,$0e     ;O
        .DB     ka5,kb1,$0f     ;P
        .DB     ka6,kb0,$10     ; (store patch)
        .DB     ka7,kb1,$11     ;<- (load patch)
        .DB     ka7,kb0,$81     ;1
        .DB     ka7,kb3,$82     ;2
        .DB     ka1,kb0,$83     ;3
        .DB     ka1,kb3,$84     ;4
        .DB     ka2,kb0,$85     ;5
        .DB     ka2,kb3,$86     ;6
        .DB     ka3,kb0,$87     ;7
        .DB     ka3,kb3,$88     ;8
        .DB     ka4,kb0,$89     ;9
        .DB     0

controlkeylist:
        .DB     ka0,kb7,$01     ;pitch bend down
        .DB     ka0,kb2,$02     ;pitch bend up
        .DB     ka0,kb4,$3c     ;F1
        .DB     ka0,kb5,$30     ;F3
        .DB     ka0,kb6,$24     ;F5
        .DB     ka0,kb3,$18     ;F7
        .DB     ka1,kb7,$80     ;left shift
        .DB     ka6,kb4,$80     ;right shift
        .DB     0

keyboardlist:
        .DB     ka1,kb4,$00     ;Z
        .DB     ka1,kb5,$01     ;S
        .DB     ka2,kb7,$02     ;X
        .DB     ka2,kb2,$03     ;D
        .DB     ka2,kb4,$04     ;C
        .DB     ka3,kb7,$05     ;V
        .DB     ka3,kb2,$06     ;G
        .DB     ka3,kb4,$07     ;B
        .DB     ka3,kb5,$08     ;H
        .DB     ka4,kb7,$09     ;N
        .DB     ka4,kb2,$0a     ;J
        .DB     ka4,kb4,$0b     ;M
        .DB     ka5,kb7,$0c     ;,
        .DB     ka7,kb6,$0c     ;Q
        .DB     ka5,kb2,$0d     ;L
        .DB     ka7,kb3,$0d     ;2
        .DB     ka5,kb4,$0e     ;.
        .DB     ka1,kb1,$0e     ;W
        .DB     ka5,kb5,$0f     ;:
        .DB     ka1,kb0,$0f     ;3
        .DB     ka6,kb7,$10     ;/
        .DB     ka1,kb6,$10     ;E
        .DB     ka2,kb1,$11     ;R
        .DB     ka2,kb0,$12     ;5
        .DB     ka2,kb6,$13     ;T
        .DB     ka2,kb3,$14     ;6
        .DB     ka3,kb1,$15     ;Y
        .DB     ka3,kb0,$16     ;7
        .DB     ka3,kb6,$17     ;U
        .DB     ka4,kb1,$18     ;I
        .DB     ka4,kb0,$19     ;9
        .DB     ka4,kb6,$1a     ;O
        .DB     ka4,kb3,$1b     ;0
        .DB     ka5,kb1,$1c     ;P
        .DB     ka5,kb6,$1d     ;@
        .DB     ka5,kb3,$1e     ;-
        .DB     ka6,kb1,$1f     ;0

parline:
        .DB     ($0400+$01*$28)&255
        .DB     ($0400+$02*$28)&255
        .DB     ($0400+$03*$28)&255
        .DB     ($0400+$06*$28)&255
        .DB     ($0400+$07*$28)&255
        .DB     ($0400+$08*$28)&255
        .DB     ($0400+$09*$28)&255
        .DB     ($0400+$0c*$28)&255
        .DB     ($0400+$0d*$28)&255
        .DB     ($0400+$0e*$28)&255
        .DB     ($0400+$0f*$28)&255
        .DB     ($0400+$12*$28)&255
        .DB     ($0400+$13*$28)&255
        .DB     ($0400+$14*$28)&255
        .DB     ($0400+$15*$28)&255
        .DB     ($0400+$16*$28)&255

parlineh:
        .DB     ($0400+$01*$28)>>8
        .DB     ($0400+$02*$28)>>8
        .DB     ($0400+$03*$28)>>8
        .DB     ($0400+$06*$28)>>8
        .DB     ($0400+$07*$28)>>8
        .DB     ($0400+$08*$28)>>8
        .DB     ($0400+$09*$28)>>8
        .DB     ($0400+$0c*$28)>>8
        .DB     ($0400+$0d*$28)>>8
        .DB     ($0400+$0e*$28)>>8
        .DB     ($0400+$0f*$28)>>8
        .DB     ($0400+$12*$28)>>8
        .DB     ($0400+$13*$28)>>8
        .DB     ($0400+$14*$28)>>8
        .DB     ($0400+$15*$28)>>8
        .DB     ($0400+$16*$28)>>8

botlinecur:
        .DB     $09,$00

flashtab:
        .DB     $00,$0b,$0c,$0f,$01,$0f,$0c,$0b

patchdata:
        .INCBIN "presets.bin"

screen:
        .INCBIN "screen.bin"
.ENDS


; crap messy code who cares feel free to do anything with it
; let me know if you add midi-in support :)

.ORG $2000
.SECTION "font" FORCE
        .INCBIN "font.bin"
.ENDS
